@@ -1,11 +1,5 @@
-//! Pure Rust implementation of Public-Key Cryptography Standards (PKCS) #5:
-//!
-//! Password-Based Cryptography Specification Version 2.1 ([RFC 8018])
-//!
-//! # Minimum Supported Rust Version
-//!
-//! This crate requires **Rust 1.51** at a minimum.
-//!
+#![doc = include_str!("../README.md")]
+
 //! # Usage
 //!
 //! The main API for this crate is the [`EncryptionScheme`] enum, which impls
@@ -20,7 +14,7 @@
 #![doc(
     html_logo_url = "https://raw.githubusercontent.com/RustCrypto/meta/master/logo.svg",
     html_favicon_url = "https://raw.githubusercontent.com/RustCrypto/meta/master/logo.svg",
-    html_root_url = "https://docs.rs/pkcs5/0.3.2"
+    html_root_url = "https://docs.rs/pkcs5/0.4.0"
 )]
 #![forbid(unsafe_code, clippy::unwrap_used)]
 #![warn(missing_docs, rust_2018_idioms, unused_qualifications)]
@@ -28,31 +22,20 @@
 #[cfg(all(feature = "alloc", feature = "pbes2"))]
 extern crate alloc;
 
+mod error;
+
 pub mod pbes1;
 pub mod pbes2;
 
-pub use der::{self, asn1::ObjectIdentifier, Error};
+pub use crate::error::{Error, Result};
+pub use der::{self, asn1::ObjectIdentifier};
 pub use spki::AlgorithmIdentifier;
 
-use core::{
-    convert::{TryFrom, TryInto},
-    fmt,
-};
-use der::{asn1::Any, message, Encodable, Encoder, Length};
+use der::{Decodable, Decoder, Encodable, Encoder, Length, Tag};
 
 #[cfg(all(feature = "alloc", feature = "pbes2"))]
 use alloc::vec::Vec;
 
-/// Cryptographic errors
-#[derive(Copy, Clone, Debug, Eq, PartialEq)]
-pub struct CryptoError;
-
-impl fmt::Display for CryptoError {
-    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
-        f.write_str("PKCS#5 cryptographic error")
-    }
-}
-
 /// Supported PKCS#5 password-based encryption schemes.
 #[derive(Clone, Debug, Eq, PartialEq)]
 #[non_exhaustive]
@@ -75,14 +58,10 @@ impl<'a> EncryptionScheme<'a> {
     #[cfg(all(feature = "alloc", feature = "pbes2"))]
     #[cfg_attr(docsrs, doc(cfg(feature = "alloc")))]
     #[cfg_attr(docsrs, doc(cfg(feature = "pbes2")))]
-    pub fn decrypt(
-        &self,
-        password: impl AsRef<[u8]>,
-        ciphertext: &[u8],
-    ) -> Result<Vec<u8>, CryptoError> {
+    pub fn decrypt(&self, password: impl AsRef<[u8]>, ciphertext: &[u8]) -> Result<Vec<u8>> {
         match self {
             Self::Pbes2(params) => params.decrypt(password, ciphertext),
-            _ => Err(CryptoError),
+            Self::Pbes1(_) => Err(Error::NoPbes1CryptSupport),
         }
     }
 
@@ -98,10 +77,10 @@ impl<'a> EncryptionScheme<'a> {
         &self,
         password: impl AsRef<[u8]>,
         buffer: &'b mut [u8],
-    ) -> Result<&'b [u8], CryptoError> {
+    ) -> Result<&'b [u8]> {
         match self {
             Self::Pbes2(params) => params.decrypt_in_place(password, buffer),
-            _ => Err(CryptoError),
+            Self::Pbes1(_) => Err(Error::NoPbes1CryptSupport),
         }
     }
 
@@ -110,14 +89,10 @@ impl<'a> EncryptionScheme<'a> {
     #[cfg(all(feature = "alloc", feature = "pbes2"))]
     #[cfg_attr(docsrs, doc(cfg(feature = "alloc")))]
     #[cfg_attr(docsrs, doc(cfg(feature = "pbes2")))]
-    pub fn encrypt(
-        &self,
-        password: impl AsRef<[u8]>,
-        plaintext: &[u8],
-    ) -> Result<Vec<u8>, CryptoError> {
+    pub fn encrypt(&self, password: impl AsRef<[u8]>, plaintext: &[u8]) -> Result<Vec<u8>> {
         match self {
             Self::Pbes2(params) => params.encrypt(password, plaintext),
-            _ => Err(CryptoError),
+            Self::Pbes1(_) => Err(Error::NoPbes1CryptSupport),
         }
     }
 
@@ -130,10 +105,10 @@ impl<'a> EncryptionScheme<'a> {
         password: impl AsRef<[u8]>,
         buffer: &'b mut [u8],
         pos: usize,
-    ) -> Result<&'b [u8], CryptoError> {
+    ) -> Result<&'b [u8]> {
         match self {
             Self::Pbes2(params) => params.encrypt_in_place(password, buffer, pos),
-            _ => Err(CryptoError),
+            Self::Pbes1(_) => Err(Error::NoPbes1CryptSupport),
         }
     }
 
@@ -162,29 +137,32 @@ impl<'a> EncryptionScheme<'a> {
     }
 }
 
-impl<'a> TryFrom<&'a [u8]> for EncryptionScheme<'a> {
-    type Error = Error;
-
-    fn try_from(bytes: &'a [u8]) -> der::Result<EncryptionScheme<'a>> {
-        AlgorithmIdentifier::try_from(bytes).and_then(TryInto::try_into)
+impl<'a> Decodable<'a> for EncryptionScheme<'a> {
+    fn decode(decoder: &mut Decoder<'a>) -> der::Result<Self> {
+        AlgorithmIdentifier::decode(decoder)?.try_into()
     }
 }
 
-impl<'a> TryFrom<Any<'a>> for EncryptionScheme<'a> {
-    type Error = Error;
-
-    fn try_from(any: Any<'a>) -> der::Result<EncryptionScheme<'a>> {
-        AlgorithmIdentifier::try_from(any).and_then(TryInto::try_into)
+impl<'a> Encodable for EncryptionScheme<'a> {
+    fn encoded_len(&self) -> der::Result<Length> {
+        match self {
+            Self::Pbes1(pbes1) => pbes1.encoded_len(),
+            Self::Pbes2(pbes2) => {
+                (pbes2::PBES2_OID.encoded_len()? + pbes2.encoded_len()?)?.for_tlv()
+            }
+        }
     }
-}
 
-impl<'a> TryFrom<AlgorithmIdentifier<'a>> for EncryptionScheme<'a> {
-    type Error = Error;
-
-    fn try_from(alg: AlgorithmIdentifier<'a>) -> der::Result<EncryptionScheme<'_>> {
-        match alg.oid {
-            pbes2::PBES2_OID => pbes2::Parameters::try_from(alg.parameters_any()?).map(Into::into),
-            _ => pbes1::Parameters::try_from(alg).map(Into::into),
+    fn encode(&self, encoder: &mut Encoder<'_>) -> der::Result<()> {
+        match self {
+            Self::Pbes1(pbes1) => pbes1.encode(encoder),
+            Self::Pbes2(pbes2) => {
+                let seq_len = (pbes2::PBES2_OID.encoded_len()? + pbes2.encoded_len()?)?;
+                encoder.sequence(seq_len, |seq| {
+                    seq.encode(&pbes2::PBES2_OID)?;
+                    seq.encode(pbes2)
+                })
+            }
         }
     }
 }
@@ -201,18 +179,26 @@ impl<'a> From<pbes2::Parameters<'a>> for EncryptionScheme<'a> {
     }
 }
 
-impl<'a> Encodable for EncryptionScheme<'a> {
-    fn encoded_len(&self) -> der::Result<Length> {
-        match self {
-            Self::Pbes1(pbes1) => pbes1.encoded_len(),
-            Self::Pbes2(pbes2) => message::encoded_len(&[&pbes2::PBES2_OID, pbes2]),
-        }
-    }
+impl<'a> TryFrom<AlgorithmIdentifier<'a>> for EncryptionScheme<'a> {
+    type Error = der::Error;
 
-    fn encode(&self, encoder: &mut Encoder<'_>) -> der::Result<()> {
-        match self {
-            Self::Pbes1(pbes1) => pbes1.encode(encoder),
-            Self::Pbes2(pbes2) => encoder.message(&[&pbes2::PBES2_OID, pbes2]),
+    fn try_from(alg: AlgorithmIdentifier<'a>) -> der::Result<EncryptionScheme<'_>> {
+        if alg.oid == pbes2::PBES2_OID {
+            if let Some(params) = alg.parameters {
+                pbes2::Parameters::try_from(params).map(Into::into)
+            } else {
+                Err(Tag::OctetString.value_error())
+            }
+        } else {
+            pbes1::Parameters::try_from(alg).map(Into::into)
         }
     }
 }
+
+impl<'a> TryFrom<&'a [u8]> for EncryptionScheme<'a> {
+    type Error = der::Error;
+
+    fn try_from(bytes: &'a [u8]) -> der::Result<EncryptionScheme<'a>> {
+        AlgorithmIdentifier::from_der(bytes)?.try_into()
+    }
+}
