@@ -2,11 +2,10 @@
 //!
 //! [RFC 8018 Section 6.1]: https://tools.ietf.org/html/rfc8018#section-6.1
 
-use crate::{AlgorithmIdentifier, Error};
-use core::convert::{TryFrom, TryInto};
+use crate::AlgorithmIdentifier;
 use der::{
-    asn1::{Any, ObjectIdentifier, OctetString},
-    message, Encodable, Encoder, ErrorKind, Header, Length, Result, Tag,
+    asn1::{ObjectIdentifier, OctetString},
+    Decodable, Decoder, Encodable, Encoder, ErrorKind, FixedTag, Length, Tag,
 };
 
 /// `pbeWithMD2AndDES-CBC` Object Identifier (OID).
@@ -65,75 +64,78 @@ impl Parameters {
     pub fn oid(&self) -> ObjectIdentifier {
         self.encryption.oid()
     }
-}
 
-impl<'a> TryFrom<Any<'a>> for Parameters {
-    type Error = Error;
+    /// Get the inner length of the encoded sequence
+    fn inner_len(&self) -> der::Result<Length> {
+        let oid_len = self.encryption.oid().encoded_len()?;
+        let params_len = (self.salt_string()?.encoded_len()?
+            + self.iteration_count.encoded_len()?)?
+        .for_tlv()?;
+        oid_len + params_len
+    }
 
-    fn try_from(any: Any<'a>) -> Result<Self> {
-        AlgorithmIdentifier::try_from(any).and_then(TryInto::try_into)
+    /// Get an [`OctetString`] wrapper for the salt
+    fn salt_string(&self) -> der::Result<OctetString<'_>> {
+        OctetString::new(&self.salt)
     }
 }
 
-impl<'a> TryFrom<AlgorithmIdentifier<'a>> for Parameters {
-    type Error = Error;
+impl Decodable<'_> for Parameters {
+    fn decode(decoder: &mut Decoder<'_>) -> der::Result<Self> {
+        AlgorithmIdentifier::decode(decoder)?.try_into()
+    }
+}
 
-    fn try_from(alg: AlgorithmIdentifier<'a>) -> Result<Self> {
-        // Ensure that we have a supported PBES1 algorithm identifier
-        let encryption = EncryptionScheme::try_from(alg.oid)
-            .map_err(|_| der::Tag::ObjectIdentifier.value_error())?;
+impl Encodable for Parameters {
+    fn encoded_len(&self) -> der::Result<Length> {
+        self.inner_len()?.for_tlv()
+    }
 
-        alg.parameters_any()?.sequence(|params| {
-            let salt = params
-                .octet_string()?
-                .as_bytes()
-                .try_into()
-                .map_err(|_| der::Tag::OctetString.value_error())?;
+    fn encode(&self, encoder: &mut Encoder<'_>) -> der::Result<()> {
+        encoder.sequence(self.inner_len()?, |encoder| {
+            encoder.oid(self.encryption.oid())?;
 
-            let iteration_count = params.decode()?;
-
-            Ok(Self {
-                encryption,
-                salt,
-                iteration_count,
+            let salt_string = self.salt_string()?;
+            let seq_len = (salt_string.encoded_len()? + self.iteration_count.encoded_len()?)?;
+            encoder.sequence(seq_len, |seq| {
+                seq.encode(&salt_string)?;
+                seq.encode(&self.iteration_count)
             })
         })
     }
 }
 
-impl Encodable for Parameters {
-    fn encoded_len(&self) -> Result<Length> {
-        self.header()?.encoded_len()
-    }
-
-    fn encode(&self, encoder: &mut Encoder<'_>) -> Result<()> {
-        self.header()?.encode(encoder)?;
-        let inner_len = self.inner_len()?;
-
-        encoder.sequence(inner_len, |nested_encoder| {
-            nested_encoder.oid(self.encryption.oid())?;
-            nested_encoder.message(&[&self.salt_string()?, &self.iteration_count])?;
-            Ok(())
-        })
-    }
+impl FixedTag for Parameters {
+    const TAG: Tag = Tag::Sequence;
 }
 
-impl Parameters {
-    /// Get the DER [`Header`]
-    fn header(&self) -> Result<Header> {
-        Header::new(Tag::Sequence, self.inner_len()?)
-    }
+impl<'a> TryFrom<AlgorithmIdentifier<'a>> for Parameters {
+    type Error = der::Error;
 
-    /// Get the inner length of the encoded sequence
-    fn inner_len(&self) -> Result<Length> {
-        let oid_len = self.encryption.oid().encoded_len()?;
-        let params_len = message::encoded_len(&[&self.salt_string()?, &self.iteration_count])?;
-        oid_len + params_len
-    }
+    fn try_from(alg: AlgorithmIdentifier<'a>) -> der::Result<Self> {
+        // Ensure that we have a supported PBES1 algorithm identifier
+        let encryption = EncryptionScheme::try_from(alg.oid)
+            .map_err(|_| der::Tag::ObjectIdentifier.value_error())?;
 
-    /// Get an [`OctetString`] wrapper for the salt
-    fn salt_string(&self) -> Result<OctetString<'_>> {
-        OctetString::new(&self.salt)
+        if let Some(any) = alg.parameters {
+            any.sequence(|params| {
+                let salt = params
+                    .octet_string()?
+                    .as_bytes()
+                    .try_into()
+                    .map_err(|_| der::Tag::OctetString.value_error())?;
+
+                let iteration_count = params.decode()?;
+
+                Ok(Self {
+                    encryption,
+                    salt,
+                    iteration_count,
+                })
+            })
+        } else {
+            Err(Tag::OctetString.value_error())
+        }
     }
 }
 
@@ -162,9 +164,9 @@ pub enum EncryptionScheme {
 }
 
 impl TryFrom<ObjectIdentifier> for EncryptionScheme {
-    type Error = Error;
+    type Error = der::Error;
 
-    fn try_from(oid: ObjectIdentifier) -> Result<Self> {
+    fn try_from(oid: ObjectIdentifier) -> der::Result<Self> {
         match oid {
             PBE_WITH_MD2_AND_DES_CBC_OID => Ok(Self::PbeWithMd2AndDesCbc),
             PBE_WITH_MD2_AND_RC2_CBC_OID => Ok(Self::PbeWithMd2AndRc2Cbc),
@@ -172,7 +174,7 @@ impl TryFrom<ObjectIdentifier> for EncryptionScheme {
             PBE_WITH_MD5_AND_RC2_CBC_OID => Ok(Self::PbeWithMd5AndRc2Cbc),
             PBE_WITH_SHA1_AND_DES_CBC_OID => Ok(Self::PbeWithSha1AndDesCbc),
             PBE_WITH_SHA1_AND_RC2_CBC_OID => Ok(Self::PbeWithSha1AndRc2Cbc),
-            _ => Err(ErrorKind::UnknownOid { oid }.into()),
+            _ => Err(ErrorKind::OidUnknown { oid }.into()),
         }
     }
 }
@@ -216,11 +218,11 @@ impl EncryptionScheme {
 }
 
 impl Encodable for EncryptionScheme {
-    fn encoded_len(&self) -> Result<Length> {
+    fn encoded_len(&self) -> der::Result<Length> {
         self.oid().encoded_len()
     }
 
-    fn encode(&self, encoder: &mut Encoder<'_>) -> Result<()> {
+    fn encode(&self, encoder: &mut Encoder<'_>) -> der::Result<()> {
         self.oid().encode(encoder)
     }
 }
