@@ -1,10 +1,9 @@
 //! Key derivation functions.
 
-use crate::{AlgorithmIdentifier, CryptoError, Error};
-use core::convert::{TryFrom, TryInto};
+use crate::{AlgorithmIdentifier, Error, Result};
 use der::{
     asn1::{Any, ObjectIdentifier, OctetString},
-    Encodable, Encoder, ErrorKind, Length, Message,
+    Decodable, Decoder, Encodable, Encoder, ErrorKind, Length, Sequence, Tag, Tagged,
 };
 
 /// Password-Based Key Derivation Function (PBKDF2) OID.
@@ -45,6 +44,14 @@ pub enum Kdf<'a> {
 }
 
 impl<'a> Kdf<'a> {
+    /// Get derived key length, if defined
+    pub fn key_length(&self) -> Option<u16> {
+        match self {
+            Self::Pbkdf2(params) => params.key_length,
+            Self::Scrypt(params) => params.key_length,
+        }
+    }
+
     /// Get the [`ObjectIdentifier`] (a.k.a OID) for this algorithm.
     pub fn oid(&self) -> ObjectIdentifier {
         match self {
@@ -78,6 +85,30 @@ impl<'a> Kdf<'a> {
     pub fn is_scrypt(&self) -> bool {
         self.scrypt().is_some()
     }
+
+    /// Convenience function to turn the OID (see [`oid`](Self::oid))
+    /// of this [`Kdf`] into error case [`Error::AlgorithmParametersInvalid`]
+    pub fn to_alg_params_invalid(&self) -> Error {
+        Error::AlgorithmParametersInvalid { oid: self.oid() }
+    }
+}
+
+impl<'a> Decodable<'a> for Kdf<'a> {
+    fn decode(decoder: &mut Decoder<'a>) -> der::Result<Self> {
+        AlgorithmIdentifier::decode(decoder)?.try_into()
+    }
+}
+
+impl<'a> Sequence<'a> for Kdf<'a> {
+    fn fields<F, T>(&self, f: F) -> der::Result<T>
+    where
+        F: FnOnce(&[&dyn Encodable]) -> der::Result<T>,
+    {
+        match self {
+            Self::Pbkdf2(params) => f(&[&self.oid(), params]),
+            Self::Scrypt(params) => f(&[&self.oid(), params]),
+        }
+    }
 }
 
 impl<'a> From<Pbkdf2Params<'a>> for Kdf<'a> {
@@ -92,40 +123,18 @@ impl<'a> From<ScryptParams<'a>> for Kdf<'a> {
     }
 }
 
-impl<'a> TryFrom<Any<'a>> for Kdf<'a> {
-    type Error = Error;
-
-    fn try_from(any: Any<'a>) -> der::Result<Self> {
-        AlgorithmIdentifier::try_from(any).and_then(TryInto::try_into)
-    }
-}
-
 impl<'a> TryFrom<AlgorithmIdentifier<'a>> for Kdf<'a> {
-    type Error = Error;
+    type Error = der::Error;
 
     fn try_from(alg: AlgorithmIdentifier<'a>) -> der::Result<Self> {
-        match alg.oid {
-            PBKDF2_OID => alg
-                .parameters_any()
-                .and_then(TryFrom::try_from)
-                .map(Self::Pbkdf2),
-            SCRYPT_OID => alg
-                .parameters_any()
-                .and_then(TryFrom::try_from)
-                .map(Self::Scrypt),
-            oid => Err(ErrorKind::UnknownOid { oid }.into()),
-        }
-    }
-}
-
-impl<'a> Message<'a> for Kdf<'a> {
-    fn fields<F, T>(&self, f: F) -> der::Result<T>
-    where
-        F: FnOnce(&[&dyn Encodable]) -> der::Result<T>,
-    {
-        match self {
-            Self::Pbkdf2(params) => f(&[&self.oid(), params]),
-            Self::Scrypt(params) => f(&[&self.oid(), params]),
+        if let Some(params) = alg.parameters {
+            match alg.oid {
+                PBKDF2_OID => params.try_into().map(Self::Pbkdf2),
+                SCRYPT_OID => params.try_into().map(Self::Scrypt),
+                oid => Err(ErrorKind::OidUnknown { oid }.into()),
+            }
+        } else {
+            Err(Tag::OctetString.value_error())
         }
     }
 }
@@ -153,7 +162,7 @@ pub struct Pbkdf2Params<'a> {
     pub salt: &'a [u8],
 
     /// PBKDF2 iteration count
-    pub iteration_count: u16,
+    pub iteration_count: u32,
 
     /// PBKDF2 output length
     pub key_length: Option<u16>,
@@ -163,8 +172,24 @@ pub struct Pbkdf2Params<'a> {
 }
 
 impl<'a> Pbkdf2Params<'a> {
+    /// Implementation defined maximum iteration count of 100,000,000.
+    ///
+    /// > For especially critical keys, or
+    /// > for very powerful systems or systems where user-perceived
+    /// > performance is not critical, an iteration count of 10,000,000 may
+    /// > be appropriate.
+    ///
+    /// See [RFC 8018, §4.2](https://datatracker.ietf.org/doc/html/rfc8018#section-4.2)
+    /// and [RFC 8018, §A.2](https://datatracker.ietf.org/doc/html/rfc8018#appendix-A.2)
+    pub const MAX_ITERATION_COUNT: u32 = 100_000_000;
+
+    const INVALID_ERR: Error = Error::AlgorithmParametersInvalid { oid: PBKDF2_OID };
+
     /// Initialize PBKDF2-SHA256 with the given iteration count and salt
-    pub fn hmac_with_sha256(iteration_count: u16, salt: &'a [u8]) -> Result<Self, CryptoError> {
+    pub fn hmac_with_sha256(iteration_count: u32, salt: &'a [u8]) -> Result<Self> {
+        if iteration_count > Self::MAX_ITERATION_COUNT {
+            return Err(Self::INVALID_ERR);
+        }
         Ok(Self {
             salt,
             iteration_count,
@@ -174,28 +199,13 @@ impl<'a> Pbkdf2Params<'a> {
     }
 }
 
-impl<'a> TryFrom<Any<'a>> for Pbkdf2Params<'a> {
-    type Error = Error;
-
-    fn try_from(any: Any<'a>) -> der::Result<Self> {
-        any.sequence(|params| {
-            // TODO(tarcieri): support salt `CHOICE` w\ `AlgorithmIdentifier`
-            let salt = params.octet_string()?;
-            let iteration_count = params.decode()?;
-            let key_length = params.optional()?;
-            let prf: Option<AlgorithmIdentifier<'_>> = params.optional()?;
-
-            Ok(Self {
-                salt: salt.as_bytes(),
-                iteration_count,
-                key_length,
-                prf: prf.map(TryInto::try_into).transpose()?.unwrap_or_default(),
-            })
-        })
+impl<'a> Decodable<'a> for Pbkdf2Params<'a> {
+    fn decode(decoder: &mut Decoder<'a>) -> der::Result<Self> {
+        decoder.any()?.try_into()
     }
 }
 
-impl<'a> Message<'a> for Pbkdf2Params<'a> {
+impl<'a> Sequence<'a> for Pbkdf2Params<'a> {
     fn fields<F, T>(&self, f: F) -> der::Result<T>
     where
         F: FnOnce(&[&dyn Encodable]) -> der::Result<T>,
@@ -217,6 +227,27 @@ impl<'a> Message<'a> for Pbkdf2Params<'a> {
     }
 }
 
+impl<'a> TryFrom<Any<'a>> for Pbkdf2Params<'a> {
+    type Error = der::Error;
+
+    fn try_from(any: Any<'a>) -> der::Result<Self> {
+        any.sequence(|params| {
+            // TODO(tarcieri): support salt `CHOICE` w\ `AlgorithmIdentifier`
+            let salt = params.octet_string()?;
+            let iteration_count = params.decode()?;
+            let key_length = params.optional()?;
+            let prf: Option<AlgorithmIdentifier<'_>> = params.optional()?;
+
+            Ok(Self {
+                salt: salt.as_bytes(),
+                iteration_count,
+                key_length,
+                prf: prf.map(TryInto::try_into).transpose()?.unwrap_or_default(),
+            })
+        })
+    }
+}
+
 /// Pseudo-random function used by PBKDF2.
 #[derive(Copy, Clone, Debug, Eq, PartialEq)]
 #[non_exhaustive]
@@ -264,16 +295,8 @@ impl Default for Pbkdf2Prf {
     }
 }
 
-impl<'a> TryFrom<Any<'a>> for Pbkdf2Prf {
-    type Error = Error;
-
-    fn try_from(any: Any<'a>) -> der::Result<Self> {
-        AlgorithmIdentifier::try_from(any).and_then(TryInto::try_into)
-    }
-}
-
 impl<'a> TryFrom<AlgorithmIdentifier<'a>> for Pbkdf2Prf {
-    type Error = Error;
+    type Error = der::Error;
 
     fn try_from(alg: AlgorithmIdentifier<'a>) -> der::Result<Self> {
         if let Some(params) = alg.parameters {
@@ -283,7 +306,7 @@ impl<'a> TryFrom<AlgorithmIdentifier<'a>> for Pbkdf2Prf {
             }
         } else {
             // TODO(tarcieri): support OPTIONAL parameters?
-            return Err(ErrorKind::Truncated.into());
+            return Err(Tag::Null.value_error());
         }
 
         match alg.oid {
@@ -292,7 +315,7 @@ impl<'a> TryFrom<AlgorithmIdentifier<'a>> for Pbkdf2Prf {
             HMAC_WITH_SHA256_OID => Ok(Self::HmacWithSha256),
             HMAC_WITH_SHA384_OID => Ok(Self::HmacWithSha384),
             HMAC_WITH_SHA512_OID => Ok(Self::HmacWithSha512),
-            oid => Err(ErrorKind::UnknownOid { oid }.into()),
+            oid => Err(ErrorKind::OidUnknown { oid }.into()),
         }
     }
 }
@@ -351,26 +374,47 @@ pub struct ScryptParams<'a> {
 }
 
 impl<'a> ScryptParams<'a> {
+    #[cfg(feature = "scrypt")]
+    const INVALID_ERR: Error = Error::AlgorithmParametersInvalid { oid: SCRYPT_OID };
+
     /// Get the [`ScryptParams`] for the provided upstream [`scrypt::Params`]
     /// and a provided salt string.
     #[cfg(feature = "scrypt")]
     #[cfg_attr(docsrs, doc(cfg(feature = "scrypt")))]
-    pub fn from_params_and_salt(
-        params: scrypt::Params,
-        salt: &'a [u8],
-    ) -> Result<Self, CryptoError> {
+    pub fn from_params_and_salt(params: scrypt::Params, salt: &'a [u8]) -> Result<Self> {
         Ok(Self {
             salt,
             cost_parameter: 1 << params.log_n(),
-            block_size: params.r().try_into().map_err(|_| CryptoError)?,
-            parallelization: params.p().try_into().map_err(|_| CryptoError)?,
+            block_size: params.r().try_into().map_err(|_| Self::INVALID_ERR)?,
+            parallelization: params.p().try_into().map_err(|_| Self::INVALID_ERR)?,
             key_length: None,
         })
     }
 }
 
+impl<'a> Decodable<'a> for ScryptParams<'a> {
+    fn decode(decoder: &mut Decoder<'a>) -> der::Result<Self> {
+        decoder.any()?.try_into()
+    }
+}
+
+impl<'a> Sequence<'a> for ScryptParams<'a> {
+    fn fields<F, T>(&self, f: F) -> der::Result<T>
+    where
+        F: FnOnce(&[&dyn Encodable]) -> der::Result<T>,
+    {
+        f(&[
+            &OctetString::new(self.salt)?,
+            &self.cost_parameter,
+            &self.block_size,
+            &self.parallelization,
+            &self.key_length,
+        ])
+    }
+}
+
 impl<'a> TryFrom<Any<'a>> for ScryptParams<'a> {
-    type Error = Error;
+    type Error = der::Error;
 
     fn try_from(any: Any<'a>) -> der::Result<Self> {
         any.sequence(|params| {
@@ -391,27 +435,12 @@ impl<'a> TryFrom<Any<'a>> for ScryptParams<'a> {
     }
 }
 
-impl<'a> Message<'a> for ScryptParams<'a> {
-    fn fields<F, T>(&self, f: F) -> der::Result<T>
-    where
-        F: FnOnce(&[&dyn Encodable]) -> der::Result<T>,
-    {
-        f(&[
-            &OctetString::new(self.salt)?,
-            &self.cost_parameter,
-            &self.block_size,
-            &self.parallelization,
-            &self.key_length,
-        ])
-    }
-}
-
 #[cfg(feature = "scrypt")]
 #[cfg_attr(docsrs, doc(cfg(feature = "scrypt")))]
 impl<'a> TryFrom<ScryptParams<'a>> for scrypt::Params {
-    type Error = CryptoError;
+    type Error = Error;
 
-    fn try_from(params: ScryptParams<'a>) -> Result<scrypt::Params, CryptoError> {
+    fn try_from(params: ScryptParams<'a>) -> Result<scrypt::Params> {
         scrypt::Params::try_from(&params)
     }
 }
@@ -419,16 +448,16 @@ impl<'a> TryFrom<ScryptParams<'a>> for scrypt::Params {
 #[cfg(feature = "scrypt")]
 #[cfg_attr(docsrs, doc(cfg(feature = "scrypt")))]
 impl<'a> TryFrom<&ScryptParams<'a>> for scrypt::Params {
-    type Error = CryptoError;
+    type Error = Error;
 
-    fn try_from(params: &ScryptParams<'a>) -> Result<scrypt::Params, CryptoError> {
+    fn try_from(params: &ScryptParams<'a>) -> Result<scrypt::Params> {
         let n = params.cost_parameter;
 
         // Compute log2 and verify its correctness
         let log_n = ((8 * core::mem::size_of::<ScryptCost>() as u32) - n.leading_zeros() - 1) as u8;
 
         if 1 << log_n != n {
-            return Err(CryptoError);
+            return Err(ScryptParams::INVALID_ERR);
         }
 
         scrypt::Params::new(
@@ -436,6 +465,6 @@ impl<'a> TryFrom<&ScryptParams<'a>> for scrypt::Params {
             params.block_size.into(),
             params.parallelization.into(),
         )
-        .map_err(|_| CryptoError)
+        .map_err(|_| ScryptParams::INVALID_ERR)
     }
 }
